/*
 * Copyright (c) 1992, by FORTE
 *
 * This software is furnished under a license and may be used,
 * copied, or disclosed only in accordance with the terms of such
 * license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or
 * otherwise made available to any other person. No title to and
 * ownership of the software is hereby transfered.
 */

#include "common.h"


typedef struct _st_irq_entry
{
	uint8_t latch;
	uint8_t mask;
	uint8_t spec_eoi;
	uint8_t ocr;
	uint8_t imr;

} IRQ_ENTRY;


#define IRQ_UNAVAIL		0x0000
#define IRQ_AVAIL		0x0001
#define IRQ_USED		0x0002

#define OCR1    0x20      // 8259-1 Operation control register
#define IMR1    0x21      // 8259-1 Mask register

#define OCR2    0xA0      // 8259-2 Operation control register
#define IMR2    0xA1      // 8259-2 Mask register


#define MAX_IRQ		16

static IRQ_ENTRY _gf1_irq[MAX_IRQ] = {
	{ 0, 0xFE, 0x60, OCR1, IMR1 },        // 0
	{ 0, 0xFD, 0x61, OCR1, IMR1 },        // 1
	{ 1, 0xFB, 0x62, OCR1, IMR1 },        // 2
	{ 3, 0xF7, 0x63, OCR1, IMR1 },        // 3
	{ 0, 0xEF, 0x64, OCR1, IMR1 },        // 4
	{ 2, 0xDF, 0x65, OCR1, IMR1 },        // 5
	{ 0, 0xBF, 0x66, OCR1, IMR1 },        // 6
	{ 4, 0x7F, 0x67, OCR1, IMR1 },        // 7
	{ 0, 0xFE, 0x60, OCR2, IMR2 },        // 8
	{ 0, 0xFD, 0x61, OCR2, IMR2 },        // 9
	{ 0, 0xFB, 0x62, OCR2, IMR2 },        // 10
	{ 5, 0xF7, 0x63, OCR2, IMR2 },        // 11
	{ 6, 0xEF, 0x64, OCR2, IMR2 },        // 12
	{ 0, 0xDF, 0x65, OCR2, IMR2 },        // 13
	{ 0, 0xBF, 0x66, OCR2, IMR2 },        // 14
	{ 7, 0x7F, 0x67, OCR2, IMR2 },        // 15
};

uint8_t get_irq_latch (int irq)
{
	return _gf1_irq[irq].latch;
}


static void __far handle_voice (void)
{
	int voice;

	uint8_t irq_source, temp1, temp2;

	uint32_t wave_ignore,
	         volume_ignore,
	         voice_bit;

	/*
	 * Clear the ignore flags. These flags are needed because
	 * we get lots of 'double' interrupts. This will only allow 1
	 * interrupt per voice
	 */

	wave_ignore   = 0;
	volume_ignore = 0;

	/*
	 * The GF1 has a fifo (sort of) of all pending wave table IRQs.
	 * You should stay here & service ALL pending waveform IRQs before
	 * returning
	 */

	while (1)
	{
		ENTER_CRITICAL;

		outportb (_gf1_data.reg_select, GET_IRQV);
		irq_source = inportb (_gf1_data.data_hi);

		LEAVE_CRITICAL;

		voice = irq_source & 0x1F;
		irq_source &= (VOICE_VOLUME_IRQ | VOICE_WAVE_IRQ);

		if (irq_source == (VOICE_VOLUME_IRQ | VOICE_WAVE_IRQ))
			break;

		voice_bit = 1 << voice;

		// See if any waveform irqs first

		if (!(irq_source & VOICE_WAVE_IRQ))
		{
			// See if we have already serviced this voice
			if (!(wave_ignore & voice_bit))
			{
				wave_ignore |= voice_bit;

				outportb (_gf1_data.voice_select, voice);
				outportb (_gf1_data.reg_select, GET_CONTROL);
				temp1 = inportb (_gf1_data.data_hi);

				outportb (_gf1_data.reg_select, GET_VOLUME_CONTROL);
				temp2 = inportb (_gf1_data.data_hi);

				/*
				 * If EITHER looping is on OR ROLLOVER is on,
				 * don't stop voice
				 */

				if (!((temp1 & VC_LOOP_ENABLE) || (temp2 & VC_ROLLOVER)))
				{
					UltraStopVoice (voice);
				}

				if (_gf1_data.wavetable_func)
					_gf1_data.wavetable_func (voice);
			}
		}

		if (!(irq_source & VOICE_VOLUME_IRQ))
		{
			if (!(volume_ignore & voice_bit))
			{
				volume_ignore |= voice_bit;

				outportb (_gf1_data.voice_select, voice);
				outportb (_gf1_data.reg_select, GET_VOLUME_CONTROL);
				temp1 = inportb (_gf1_data.data_hi);

				// If volume looping is enabled, don't stop it
				if (!(temp1 & VL_LOOP_ENABLE))
				{
					UltraStopVolume (voice);
				}

				if (_gf1_data.volume_func)
					_gf1_data.volume_func (voice);
			}
		}
	}

	return;
}


static void __far gf1_handler (void)
{
	/*
	 * IRQ's that will be vectored here
	 * --------------------------------
	 * 1) DAC sampling (PC DMA)
	 * 2) Voice Volume Envelope
	 * 3) WAVE table end
	 * 4) DMA to DRAM
	 * 5) All MIDI functions
	 */

	while (1)
	{
		uint8_t irq_source = inportb (_gf1_data.irq_status);

		if (irq_source == 0)
			break;

		if (irq_source & DMA_TC_IRQ)
			handle_dma_tc ();

		if (irq_source & (MIDI_TX_IRQ | MIDI_RX_IRQ))
		{
			uint8_t midi_status = inportb (_gf1_data.midi_control);

			if (irq_source & MIDI_TX_IRQ)
			{
				if (_gf1_data.midi_xmit_func)
					_gf1_data.midi_xmit_func (midi_status);
			}

			if (irq_source & MIDI_RX_IRQ)
			{
				// reading data will clear IRQ
				uint8_t data = inportb (_gf1_data.midi_data);

				if (_gf1_data.midi_recv_func)
					_gf1_data.midi_recv_func (midi_status, data);
			}
		}

		if (irq_source & GF1_TIMER1_IRQ)
		{
			ENTER_CRITICAL;

			outportb (_gf1_data.reg_select, TIMER_CONTROL);
			outportb (_gf1_data.data_hi, _gf1_data.timer_ctrl & ~0x04);

			outportb (_gf1_data.reg_select, TIMER_CONTROL);
			outportb (_gf1_data.data_hi, _gf1_data.timer_ctrl);

			if (_gf1_data.timer1_func)
				_gf1_data.timer1_func ();

			LEAVE_CRITICAL;
		}

		if (irq_source & GF1_TIMER2_IRQ)
		{
			ENTER_CRITICAL;

			outportb (_gf1_data.reg_select, TIMER_CONTROL);
			outportb (_gf1_data.data_hi, _gf1_data.timer_ctrl & ~0x08);

			outportb (_gf1_data.reg_select, TIMER_CONTROL);
			outportb (_gf1_data.data_hi, _gf1_data.timer_ctrl);

			if (_gf1_data.timer2_func)
				_gf1_data.timer2_func ();

			LEAVE_CRITICAL;
		}

		if (irq_source & (WAVETABLE_IRQ | ENVELOPE_IRQ))
			handle_voice ();
	}

	if (_gf1_data.aux_irq_func)
		_gf1_data.aux_irq_func ();

	return;
}

//static void (__interrupt __far *prev_handler_gf1)();


void __interrupt __far gf1_irq_handler (void)
{
	int irq_num = _gf1_data.gf1_irq_num;

	outportb (_gf1_irq[irq_num].ocr, _gf1_irq[irq_num].spec_eoi);

	if (irq_num > 7)
		outportb (OCR1, EOI);

	gf1_handler ();

//	_chain_intr (prev_handler_gf1);

	return;
}

#if 0
void __interrupt __far midi_irq_handler (void)
{
	int irq_num = _gf1_data.midi_irq_num;

	outportb (_gf1_irq[irq_num].ocr, _gf1_irq[irq_num].spec_eoi);

	if (irq_num > 7)
		outportb (OCR1, EOI);

	gf1_handler ();

	return;
}
#endif





/***************************************************************************/
/* INTERRUPT HOOK FUNCTIONS ************************************************/
/***************************************************************************/

static inline int ivect (int irq)
{
	return (irq < 8) ? irq + 8 : irq + 0x68;
}


static void (__interrupt __far *prev_handler_gf1)();

static void gf1_hook (int irq)
{
	prev_handler_gf1 = _dos_getvect (ivect (irq));
	_dos_setvect (ivect (irq), gf1_irq_handler);

	return;
}


static void gf1_unhook (int irq)
{
	_dos_setvect (ivect (irq), prev_handler_gf1);

	return;
}


static void midi_hook (int irq)
{
	return;
}


static void midi_unhook (int irq)
{
	return;
}


void ReSetIrqHandlers (int gf1_irq, int midi_irq)
{
	if (gf1_irq != 0)
		gf1_unhook (gf1_irq);

	if ((gf1_irq != midi_irq) && (midi_irq != 0))
		midi_unhook (midi_irq);

	return;
}


void SetIrqHandlers (int gf1_irq, int midi_irq)
{
	if (gf1_irq != 0)
		gf1_hook (gf1_irq);

	if ((midi_irq != 0) && (midi_irq != gf1_irq))
		midi_hook (midi_irq);

	return;
}


void SetIrqs (int gf1_irq, int midi_irq)
{
	uint8_t val;

	if (gf1_irq != 0)
	{
		// unmask gf1 interrupt
		val = inportb (_gf1_irq[gf1_irq].imr);
		val &= _gf1_irq[gf1_irq].mask;
		outportb (_gf1_irq[gf1_irq].imr, val);

		// send a specific EOI in case of pending interrupt
		outportb (_gf1_irq[gf1_irq].ocr, _gf1_irq[gf1_irq].spec_eoi);
	}

	if ((midi_irq != gf1_irq) && (midi_irq != 0))
	{
		// unmask midi interrupt
		val = inportb (_gf1_irq[midi_irq].imr);
		val &= _gf1_irq[midi_irq].mask;
		outportb (_gf1_irq[midi_irq].imr, val);

		// send a specific EOI in case of pending interrupt
		outportb (_gf1_irq[midi_irq].ocr, _gf1_irq[midi_irq].spec_eoi);
	}

	// Un-mask IRQ 2 from first controller if using 2nd controller
	if (midi_irq > 7 || gf1_irq > 7)
	{
		val = inportb (_gf1_irq[2].imr);
		val &= _gf1_irq[2].mask;
		outportb (_gf1_irq[2].imr, val);

		// send a specific EOI in case of pending interrupt
		outportb (_gf1_irq[2].ocr, _gf1_irq[2].spec_eoi);
	}

	return;
}


void ResetIrqs (int gf1_irq, int midi_irq)
{
	uint8_t val;

	// unmask gf1 interrupt
	if ((gf1_irq != 2) && (gf1_irq != 0))
	{
		// turn mask bit back on ...
		val = inportb (_gf1_irq[gf1_irq].imr);
		val |= ~_gf1_irq[gf1_irq].mask;
		outportb (_gf1_irq[gf1_irq].imr, val);
	}

	if ((midi_irq != 2) && (midi_irq != 0))
	{
		// unmask midi interrupt
		val = inportb (_gf1_irq[midi_irq].imr);
		val |= ~_gf1_irq[midi_irq].mask;
		outportb (_gf1_irq[midi_irq].imr, val);
	}

	return;
}
