/*
 * Copyright (c) 1992, by FORTE
 *
 * This software is furnished under a license and may be used,
 * copied, or disclosed only in accordance with the terms of such
 * license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or
 * otherwise made available to any other person. No title to and
 * ownership of the software is hereby transfered.
 */

#include "common.h"

#define POOL_SIZE (uint32_t) ((256L * 1024L) - 32L)

#define NEXT_OFFSET		0UL
#define PREV_OFFSET		4UL
#define SIZE_OFFSET		8UL

#define MEM_HEADER_SIZE SIZE_OFFSET + 4UL


uint8_t UltraPeekData (uint16_t port, uint32_t address)
{
	uint8_t ret;

	ENTER_CRITICAL;

	// Send the address where the read will be from to the GUS
	outportb (port + GF1_REG_SELECT, SET_DRAM_LOW);
	outportw (port + GF1_DATA_LOW, address & 0xffff);
	outportb (port + GF1_REG_SELECT, SET_DRAM_HIGH);
	outportb (port + GF1_DATA_HI, (address >> 16) & 0xff);
     
	ret = inportb (port + GF1_DRAM);
     
	LEAVE_CRITICAL;
     
	return ret;
}


void UltraPokeData (uint16_t port, uint32_t address, uint8_t data)
{
	ENTER_CRITICAL;

	// Send the address where the read will be from to the GUS
	outportb (port + GF1_REG_SELECT, SET_DRAM_LOW);
	outportw (port + GF1_DATA_LOW, address & 0xffff);
	outportb (port + GF1_REG_SELECT, SET_DRAM_HIGH);
	outportb (port + GF1_DATA_HI, (address >> 16) & 0xff);

	outportb (port + GF1_DRAM, data);

	LEAVE_CRITICAL;

	return;
}


void UltraPokeLong (uint32_t location, uint32_t value)
{
	uint8_t *ptr;
	int i;

	ptr = (uint8_t *) &value;

     for (i = 0; i < 4; i++)
		 UltraPokeData (_gf1_data.base_port, location++, *ptr++);
}


uint32_t UltraPeekLong (uint32_t location)
{
	uint32_t ret = 0;
	uint8_t *ptr;
	int i;

	ptr = (uint8_t *) &ret;

	for (i = 0; i < 4; i++)
		*ptr++ = UltraPeekData (_gf1_data.base_port, location++);

	return ret;
}


int UltraMemInit (void)
{
	uint32_t pool_len;
	uint32_t location;
	int size;

	if (_gf1_data.reserved_dram > (uint32_t) (256L * 1024L))
		_gf1_data.reserved_dram = 32;

	size = UltraSizeDram ();

	pool_len = (((uint32_t) size * 1024UL) - 32UL) - _gf1_data.reserved_dram;

	_gf1_data.free_mem = _gf1_data.reserved_dram;

	UltraPokeLong (_gf1_data.reserved_dram + NEXT_OFFSET, 0);
	UltraPokeLong (_gf1_data.reserved_dram + PREV_OFFSET, 0);
	UltraPokeLong (_gf1_data.reserved_dram + SIZE_OFFSET, pool_len);

	/*
	 * This section is necessary to split the DRAM into pools that
	 * do not cross 256K boundaries. the GF1 cannot play 16 bit data
	 * across 256K boundaries
	 */

	if (pool_len > (uint32_t) (256L * 1024L))
	{
		UltraMemAlloc (pool_len, &location);

		if (pool_len > (uint32_t) (256L * 3L * 1024L))
		{
			UltraMemFree (POOL_SIZE, (uint32_t) (256L * 3L * 1024L) + 32L);
		}
		if (pool_len > (uint32_t) (256L * 2L * 1024L))
		{
			UltraMemFree (POOL_SIZE, (uint32_t) (256L * 2L * 1024L) + 32L);
		}

		UltraMemFree (POOL_SIZE, (uint32_t) (256L * 1L * 1024L) + 32L);
		UltraMemFree ((uint32_t) (256L * 1024L) - _gf1_data.reserved_dram,
		              _gf1_data.reserved_dram);
	}

	return size;
}


uint32_t UltraMaxAlloc (void)
{
	uint32_t size = 0L;
	uint32_t pool_size;
	uint32_t ptr = _gf1_data.free_mem;

	while (ptr)
	{
		pool_size = UltraPeekLong (ptr + SIZE_OFFSET);

		if (pool_size > size)
			size = pool_size;

		ptr = UltraPeekLong (ptr + NEXT_OFFSET);
	}

	return size;
}


int UltraMemAlloc (uint32_t size, uint32_t *location)
{
	uint32_t ptr;
	uint32_t pool_size;
	uint32_t size_left;
	uint32_t prev;
	uint32_t next;

	// Round size up to next 32 byte boundary
	size = (size + 31UL) & -31L;

	ptr = _gf1_data.free_mem;

	while (ptr)
	{
		// See if this buffer will fit
		pool_size = UltraPeekLong (ptr + SIZE_OFFSET);

		if (pool_size >= size)
		{
			size_left = pool_size - size;

			if (size_left < MEM_HEADER_SIZE)
			{
				next = UltraPeekLong (ptr + NEXT_OFFSET);
				prev = UltraPeekLong (ptr + PREV_OFFSET);

				if (next)
					UltraPokeLong (next + PREV_OFFSET, prev);

				if (prev)
					UltraPokeLong (prev + NEXT_OFFSET, next);
				else
					_gf1_data.free_mem = next;
			
				*location = ptr;
			}
			else
			{
				// adjust size of this pool
				UltraPokeLong (ptr + SIZE_OFFSET, pool_size - size);

				*location = ptr + (pool_size - size);
			}

			return ULTRA_OK;
		}

		ptr = UltraPeekLong (ptr + NEXT_OFFSET);   // next
	}

	return NO_MEMORY;
}


static void combine_mem (void)
{
	uint32_t ptr;
	uint32_t next;
	uint32_t next_next;
	uint32_t pool_size;
	uint32_t next_size;

	ptr = _gf1_data.free_mem;

	while (ptr != 0)
	{
		next = UltraPeekLong (ptr + NEXT_OFFSET);
		pool_size = UltraPeekLong (ptr + SIZE_OFFSET);

		if ((ptr + pool_size) == next)
		{
			next_next = UltraPeekLong (next + NEXT_OFFSET);
			next_size = UltraPeekLong (next + SIZE_OFFSET);
			pool_size += next_size;
			UltraPokeLong (ptr + SIZE_OFFSET, pool_size);
			UltraPokeLong (ptr + NEXT_OFFSET, next_next);
	       
			if (next_next != 0)
				UltraPokeLong (next_next + PREV_OFFSET, ptr);
			else
				ptr = 0;   // end
		}
		else
			ptr = UltraPeekLong (ptr + NEXT_OFFSET);   // next
	}
}


int UltraMemFree (uint32_t size, uint32_t location)
{
	uint32_t ptr;
	uint32_t prev;
	uint32_t next;
	int flag = 0;

	// Round size up to next 32 byte boundary
	size = (size + 31) & -31L;

	ptr = _gf1_data.free_mem;

	if (ptr == 0)
	{
		// No free ram. Make this the first in the list
		_gf1_data.free_mem = location;

		UltraPokeLong (location + NEXT_OFFSET, 0);
		UltraPokeLong (location + PREV_OFFSET, 0);
		UltraPokeLong (location + SIZE_OFFSET, size);

		flag = 1;
	}
	else
	{
		while (ptr && !flag)
		{
			next = UltraPeekLong (ptr + NEXT_OFFSET);
			prev = UltraPeekLong (ptr + PREV_OFFSET);

			if (location < ptr)
			{
				if (prev == 0)
					_gf1_data.free_mem = location;
				else
					UltraPokeLong (prev + NEXT_OFFSET, location);

				UltraPokeLong (location + NEXT_OFFSET, ptr);
				UltraPokeLong (location + PREV_OFFSET, prev);
				UltraPokeLong (location + SIZE_OFFSET, size);
				UltraPokeLong (ptr + PREV_OFFSET, location);

				flag = 1;
			}
			else
			{
				if (next == 0)
				{
					UltraPokeLong (ptr + NEXT_OFFSET, location);
					UltraPokeLong (location + NEXT_OFFSET, 0);
					UltraPokeLong (location + PREV_OFFSET, ptr);
					UltraPokeLong (location + SIZE_OFFSET, size);

					flag = 1;
				}
			}

			ptr = UltraPeekLong (ptr + NEXT_OFFSET);   // next
		}
	}

	if (flag)
	{
		combine_mem();

		return ULTRA_OK;
	}

	return CORRUPT_MEM;
}
