/*
 * Copyright (c) 1992, by FORTE
 *
 * This software is furnished under a license and may be used,
 * copied, or disclosed only in accordance with the terms of such
 * license and with the inclusion of the above copyright notice.
 * This software or any other copies thereof may not be provided or
 * otherwise made available to any other person. No title to and
 * ownership of the software is hereby transfered.
 */

#include "common.h"


void UltraClearVoices (void)
{
	_gf1_data.used_voices = 0;

	return;
}


int UltraAllocVoice (int voice_num, int *new_num)
{
	uint32_t i;

	if (voice_num >= _gf1_data.voices)
		return VOICE_OUT_OF_RANGE;

	if (voice_num != -1)
	{
		if (!(_gf1_data.used_voices & (1L << voice_num)))
		{
			;   // do nothing, allocate it below
		}
		else
		{
			return VOICE_NOT_FREE;
		}
	}
	else
	{
		voice_num = 99;

		for (i = 0; i < (uint32_t) _gf1_data.voices; i++)
		{
			if (!(_gf1_data.used_voices & (1L << i)))
			{
				voice_num = i;
				break;
			}
		}
	}

	if (voice_num >= _gf1_data.voices)
		return NO_FREE_VOICES;

	_gf1_data.used_voices |= (1L << voice_num);

	*new_num = voice_num;

	return ULTRA_OK;
}


void UltraFreeVoice (int voice_num)
{
	_gf1_data.used_voices &= ~(1L << voice_num);
}


void UltraSetVoiceEnd (int voice, uint32_t end)
{
	uint32_t phys_end;
	uint8_t data;

	ENTER_CRITICAL;

	outportb (_gf1_data.voice_select, voice);

	outportb (_gf1_data.reg_select, GET_CONTROL);
	data = inportb (_gf1_data.data_hi);

	if (data & VC_DATA_TYPE)
		phys_end = convert_to_16bit (end);
	else
		phys_end = end;

	/* Set end address of buffer */
	outportb (_gf1_data.reg_select, SET_END_HIGH);
	outportw (_gf1_data.data_low, ADDR_HIGH(phys_end));
	outportb (_gf1_data.reg_select, SET_END_LOW);
	outportw (_gf1_data.data_low, ADDR_LOW(phys_end));
	
	// turn 'stop' bits off (prob are)
	data &= ~(VC_IRQ_PENDING | VOICE_STOPPED | STOP_VOICE);

	outportb (_gf1_data.reg_select,SET_CONTROL);
	outportb (_gf1_data.data_hi,data);
     
	gf1_delay ();
     
	outportb (_gf1_data.reg_select,SET_CONTROL);
	outportb (_gf1_data.data_hi,data);

	LEAVE_CRITICAL;

	return;
}


/*
 *   The formula for this table is:
 *        1,000,000 / (1.619695497 * # of active voices)
 *
 *        The 1.619695497 is calculated by knowing that 14 voices
 *                gives exactly 44.1 Khz. Therefore,
 *                1,000,000 / (X * 14) = 44100
 *                X = 1.619695497
 
 
 Wrong.  It's actually the base freq (9878400) / (16 * number of voices (lower cap at 14))
 The GUS needs 16 cycles per active channel with a max sample rate of 44.1 (14 voices)
 */

static uint32_t freq_divisor[19] = {
	44100,    // 14 active voices
	41160,    // 15 active voices
	38587,    // 16 active voices
	36317,    // 17 active voices
	34300,    // 18 active voices
	32494,    // 19 active voices
	30870,    // 20 active voices
	29400,    // 21 active voices
	28063,    // 22 active voices
	26843,    // 23 active voices
	25725,    // 24 active voices
	24696,    // 25 active voices
	23746,    // 26 active voices
	22866,    // 27 active voices
	22050,    // 28 active voices
	21289,    // 29 active voices
	20580,    // 30 active voices
	19916,    // 31 active voices
	19293     // 32 active voices
};

void UltraSetFrequency (int voice, uint32_t speed_khz)
{
	uint32_t fc, temp;

	// FC is calculated based on the # of active voices ...
	temp = freq_divisor[_gf1_data.voices - 14];

	fc = ((speed_khz << 9L) + (temp >> 1L)) / temp;
	fc = fc << 1;

	ENTER_CRITICAL;

	outportb (_gf1_data.voice_select, voice);
	outportb (_gf1_data.reg_select, SET_FREQUENCY);
	outportw (_gf1_data.data_low, fc);

	LEAVE_CRITICAL;

	return;
}


void UltraSetLoopMode (int voice, uint8_t mode)
{
	uint8_t data, vmode;

	ENTER_CRITICAL;

	outportb (_gf1_data.voice_select, voice);

	// set/reset the rollover bit as per user request
	outportb (_gf1_data.reg_select, GET_VOLUME_CONTROL);

	vmode = inportb (_gf1_data.data_hi);

	if (mode & USE_ROLLOVER)
	{
		vmode |= VC_ROLLOVER;
	}
	else
	{
		vmode &= ~VC_ROLLOVER;
	}

	outportb (_gf1_data.reg_select, SET_VOLUME_CONTROL);
	outportb (_gf1_data.data_hi, vmode);

	gf1_delay ();

	outportb (_gf1_data.data_hi, vmode);

	outportb (_gf1_data.reg_select, GET_CONTROL);
	data = inportb (_gf1_data.data_hi);

	data &= ~(VC_WAVE_IRQ | VC_BI_LOOP | VC_LOOP_ENABLE);
	mode &=   VC_WAVE_IRQ | VC_BI_LOOP | VC_LOOP_ENABLE;
	data |= mode;

	outportb (_gf1_data.reg_select, SET_CONTROL);
	outportb (_gf1_data.data_hi, data);

	gf1_delay ();

	outportb (_gf1_data.reg_select, SET_CONTROL);
	outportb (_gf1_data.data_hi, data);

	LEAVE_CRITICAL;
}


/*
 * This function will read the accumulator for a particular 
 * voice.
 */
uint32_t UltraReadVoice (int voice)
{
	uint16_t count_low, count_high;
	uint32_t acc;
	uint8_t mode;

	ENTER_CRITICAL;

	outportb (_gf1_data.voice_select, voice);
	
	// Get the high portion of the accumulator
	outportb (_gf1_data.reg_select, GET_ACC_HIGH);		
	count_high = inportw (_gf1_data.data_low);
	
	// Get the low portion of the accumulator
	outportb (_gf1_data.reg_select, GET_ACC_LOW);
	count_low = inportw (_gf1_data.data_low);
	
	// convert from UltraSound's format to a physical address

	outportb (_gf1_data.reg_select, GET_CONTROL);
	mode = inportb (_gf1_data.data_hi);

	LEAVE_CRITICAL;
	
	acc = make_physical_address (count_low, count_high, mode);
	acc &= 0xfffffL;		// Only 20 bits

	return acc;
}


void UltraSetVoice (int voice,uint32_t location)
{
	uint32_t phys_loc;
	uint8_t data;

	ENTER_CRITICAL;

	outportb (_gf1_data.voice_select, voice);

	outportb (_gf1_data.reg_select, GET_CONTROL);
	data = inportb (_gf1_data.data_hi);

	if (data & VC_DATA_TYPE)
		phys_loc = convert_to_16bit (location);
	else
		phys_loc = location;

	// First set accumulator to beginning of data
	outportb (_gf1_data.reg_select, SET_ACC_HIGH);
	outportw (_gf1_data.data_low, ADDR_HIGH(phys_loc));
	outportb (_gf1_data.reg_select, SET_ACC_LOW);
	outportw (_gf1_data.data_low, ADDR_LOW(phys_loc));

	LEAVE_CRITICAL;
}


uint8_t UltraPrimeVoice (int voice, uint32_t begin, uint32_t start, uint32_t end, uint8_t mode)
{
	uint32_t phys_start, phys_end;
	uint32_t phys_begin;
	uint8_t vmode;

	/* if start is greater than end, flip 'em and turn on */
	/* decrementing addresses */
	if (start > end)
	{
		uint32_t temp = start;
		start = end;
		end = temp;
		mode |= VC_DIRECT;
	}

	/* if 16 bit data, must convert addresses */
	if (mode & VC_DATA_TYPE)
	{
		phys_begin = convert_to_16bit(begin);
		phys_start = convert_to_16bit(start);
		phys_end   = convert_to_16bit(end);
	}
	else
	{
		phys_begin = begin;
		phys_start = start;
		phys_end = end;
	}

	ENTER_CRITICAL;

	outportb (_gf1_data.voice_select, voice);

	// set/reset the rollover bit as per user request
	outportb (_gf1_data.reg_select, GET_VOLUME_CONTROL);
	vmode = inportb (_gf1_data.data_hi);

	if (mode & USE_ROLLOVER)
		vmode |= VC_ROLLOVER;
	else
		vmode &= ~VC_ROLLOVER;

	outportb (_gf1_data.reg_select, SET_VOLUME_CONTROL);
	outportb (_gf1_data.data_hi, vmode);

	gf1_delay ();

	outportb (_gf1_data.data_hi, vmode);

	// First set accumulator to beginning of data
	outportb (_gf1_data.reg_select, SET_ACC_HIGH);
	outportw (_gf1_data.data_low, ADDR_HIGH(phys_begin));
	outportb (_gf1_data.reg_select, SET_ACC_LOW);
	outportw (_gf1_data.data_low, ADDR_LOW(phys_begin));

	// Set start loop address of buffer
	outportb (_gf1_data.reg_select, SET_START_HIGH);
	outportw (_gf1_data.data_low, ADDR_HIGH(phys_start));
	outportb (_gf1_data.reg_select, SET_START_LOW);
	outportw (_gf1_data.data_low, ADDR_LOW(phys_start));

	// Set end address of buffer
	outportb (_gf1_data.reg_select, SET_END_HIGH);
	outportw (_gf1_data.data_low, ADDR_HIGH(phys_end));
	outportb (_gf1_data.reg_select, SET_END_LOW);
	outportw (_gf1_data.data_low, ADDR_LOW(phys_end));

	LEAVE_CRITICAL;

	return mode;
}


void UltraGoVoice (int voice, uint8_t mode)
{
	ENTER_CRITICAL;

	outportb (_gf1_data.voice_select, voice);

	mode &= ~(VOICE_STOPPED | STOP_VOICE);    // turn 'stop' bits off

	/* NOTE: no irq's from the voice ... */
	outportb (_gf1_data.reg_select, SET_CONTROL);
	outportb (_gf1_data.data_hi, mode);

	gf1_delay ();

	outportb (_gf1_data.data_hi, mode);

	LEAVE_CRITICAL;

	return;
}


/*
 * This function will start playing a wave out of DRAM. It assumes
 * the playback rate, volume & balance have been set up before ...
 */
void UltraStartVoice (int voice, uint32_t begin, uint32_t start, uint32_t end, uint8_t mode)
{
	mode = UltraPrimeVoice (voice, begin, start, end, mode);
	UltraGoVoice (voice, mode);
}


/*
 * This function will stop a given voices output. Note that a delay
 * is necessary after the stop is issued to ensure the self-
 * modifying bits aren't a problem.
 */
void UltraStopVoice (int voice)
{
	uint8_t data;

	ENTER_CRITICAL;

	outportb (_gf1_data.voice_select, voice);

	// turn off the roll over bit first ...
	outportb (_gf1_data.reg_select, GET_VOLUME_CONTROL);

	data = inportb (_gf1_data.data_hi);
	data &= ~VC_ROLLOVER;

	outportb (_gf1_data.reg_select, SET_VOLUME_CONTROL);
	outportb (_gf1_data.data_hi, data);     // turn it off

	gf1_delay ();

	outportb (_gf1_data.data_hi, data);     // turn it off

	// Now stop the voice
	outportb (_gf1_data.reg_select, GET_CONTROL);
	data = inportb (_gf1_data.data_hi);

	data &= ~VC_WAVE_IRQ;        // disable irq's & stop voice ..
	data |= VOICE_STOPPED | STOP_VOICE;

	outportb (_gf1_data.reg_select, SET_CONTROL);
	outportb (_gf1_data.data_hi, data);     // turn it off

	gf1_delay ();

	outportb (_gf1_data.data_hi, data);     // turn it off

	LEAVE_CRITICAL;
}


int UltraVoiceStopped (int voice)
{
	uint8_t mode;

	ENTER_CRITICAL;

	// Make sure was are talking to proper voice
	outportb (_gf1_data.voice_select, voice);
	outportb (_gf1_data.reg_select, GET_CONTROL);

	mode = inportb (_gf1_data.data_hi);

	LEAVE_CRITICAL;

	return mode & (VOICE_STOPPED | STOP_VOICE);
}

/*
 * NOTE: To play a sample backwards, 'begin' should be set to address
 * that you want it to start at, start must be a lower address than the
 * end. Also turn on the decreasing addresses bit in the control byte.
 * Now the sample will begin playing at begin, with decreasing addresses
 * thru the end location until it hits the start location. It will then
 * loop back to the end position, (if looping enabled)
 */
void UltraVoiceOn (int voice, uint32_t begin, uint32_t start_loop, uint32_t end_loop, uint8_t control, uint32_t freq)
{
	// Might want to do a noteoff() here, just in case ....

	UltraSetFrequency (voice, freq);
	UltraStartVoice (voice, begin, start_loop, end_loop, control);
}


void UltraVoiceOff (int voice, int end)
{
	if (end)
	{
		/*
		 * This could be used to change the end point and let
		 * the sample stop at end (sampled decay.... )
		 */
		UltraSetLoopMode (voice, 0);
	}
	else
	{
		UltraStopVoice (voice);
	}
}

